
/*
 * $Id: Connection.cc,v 1.50 1999/09/22 21:59:46 rousskov Exp $
 *
 *
 * Web Polygraph                        http://polygraph.ircache.net/
 * ------------------------------------------------------------------
 *
 *  Web Polygraph is copyrighted (C) 1999 by Alex Rousskov and the
 *  University of California, San Diego (UCSD); all rights reserved.
 *
 *  You may copy and distribute VERBATIM copies of this product. 
 *  You may modify your copy for your own use. You must NOT
 *  distribute modified copies. You must NOT publish benchmarking
 *  results based on modified versions.
 *
 *  If you publish any materials derived from or related to Web
 *  Polygraph, you must give a credit reference to the IRCache group.
 *
 *  Web Polygraph is distributed without any warranty, even implied.
 *
 *  Please see the COPYRIGHT file for full details.
 *
 */

#include "polygraph.h"

#include <netinet/in.h>
#include <arpa/inet.h>

#include "Clock.h"
#include "PortMgr.h"
#include "ErrorMgr.h"
#include "StatPhase.h"
#include "StatPhaseMgr.h"
#include "Connection.h"

#include "polyBcastChannels.h"
#include "globals.h"


/* Connection::HalfPipe */

void Connection::HalfPipe::reset() {
	Assert(!theReserv);
	theBuf.reset();
	theIOCnt = 0;
	isReady = false;
}

void Connection::HalfPipe::start(Select::User *u) {
	Assert(!theReserv);
	theReserv = TheFileScanner->setFD(theConn.theSock.fd(), theDir, u);
}

void Connection::HalfPipe::stop(Select::User *) {
	if (theReserv)
		TheFileScanner->clearRes(theReserv);
}


/* Connection */

Connection::Connection():
	theRd(*this, theRdBuf, dirRead),
	theWr(*this, theWrBuf, dirWrite) {
	reset();
}

void Connection::reset() {
	if (theSock)
		close();
	theAddr = NetAddr();
	theMgr = 0;
	thePortMgr = 0;
	theRd.reset();
	theWr.reset();
	theCloseKind = ConnCloseStat::ckActive;
	isBad = isAtEof = isLastUse = false;
	theLocPort = theRemPort = -1;

	theUseLmt = -1; // unlimited use by default
	theUseCnt = 0;
	theOpenTime = theUseStart = Time();
}

int Connection::rport() const {
	if (!theSock)
		return -1;
	else
	if (theRemPort >= 0)
		return theRemPort;
	else
		return theRemPort = theSock.rport(); // cache the result
}

bool Connection::connect(const NetAddr &addr, const SockOpt &opt, PortMgr *aPortMgr) {
	Assert(aPortMgr);
	Assert(!theSock);
	if (!theSock.create()) {
		ReportError(errno);
		return false;
	}

	setSockOpt(opt);

	thePortMgr = aPortMgr;
	theLocPort = thePortMgr->bind(theSock);
	if (theLocPort < 0)
		return false;

	theOpenTime = TheClock;
	Broadcast(TheConnOpenChannel, this);

	return theSock.connect(theAddr = addr);
}

bool Connection::accept(Socket &s, const SockOpt &opt, bool &fatal) {
	Assert(!theSock);

	sockaddr_in addr;
	socklen_t addr_len = sizeof(addr);
	theSock = s.accept((struct sockaddr*)&addr, &addr_len);
	if (!theSock) {
		fatal = errno != EWOULDBLOCK && errno != EAGAIN;
		if (fatal)
			ReportError(errno);
		return false;
	}

	setSockOpt(opt);
	theAddr.addr(addr.sin_addr);
	theAddr.port(ntohs(addr.sin_port));
	theLocPort = -1; // unknown, do not care;
	theOpenTime = TheClock;

	Broadcast(TheConnOpenChannel, this);

	return true;
}

Size Connection::read() {
	if (!theRdBuf.spaceSize()) {
		theRd.isReady = true;
		return 0;
	}
	theRd.isReady = false;

	const Size sz = preIO(theRd) ?
		theSock.read(theRdBuf.space(), theRdBuf.spaceSize()) : (Size)-1;
	
	//cerr << here << "read " << sz << " : {" << theRdBuf.space() << "}" << endl;

	if (sz > 0)
		theRdBuf.appended(sz);
	else
	if (sz == 0)
		isAtEof = true;
	else
	if (sz < 0) {
		ReportError(errno);
		isBad = true;
	}

	if (sz >= 0)
		TheStatPhaseMgr->noteSockRead(sz);
	return sz;
}

Size Connection::write() {
	if (!theWrBuf.contSize()) {
		theWr.isReady = true;
		return 0;
	}
	theWr.isReady = false;

	const Size sz = preIO(theWr) ?
		theSock.write(theWrBuf.content(), theWrBuf.contSize()) : (Size)-1;
	
	//cerr << here << "wrote " << sz << " : {" << theWrBuf.content() << "}" << endl;

	if (sz > 0)
		theWrBuf.consumed(sz);
	else
	if (sz < 0) {
		ReportError(errno);
		isBad = true;
	}

	if (sz >= 0)
		TheStatPhaseMgr->noteSockWrite(sz);

	return sz;
}

bool Connection::close() {
	if (theLocPort >= 0) {
		thePortMgr->release(theLocPort, true);
		theLocPort = -1;
	}
	if (theSock) {
		theRd.stop(0);
		theWr.stop(0);
		Broadcast(TheConnCloseChannel, this);
		if (!theSock.close()) {
			ReportError(errno);
			return false;
		}
	}
	return true;
}

// this has to be done for accepted sockets as well
// FreeBSD accept(2) man page misleads
void Connection::setSockOpt(const SockOpt &opt) {
	Must(theSock.blocking(false));
	Must(theSock.configure(opt));
}

bool Connection::preIO(HalfPipe &ioPipe) {
	if (theRd.theIOCnt + theWr.theIOCnt == 0) {
		// check for connect/accept errors
		if (theSock.errNo())
			return false;
		Broadcast(TheConnEstChannel, this);
	}
	ioPipe.theIOCnt++;
	return true;
}

bool Connection::reusable() const {
	if (exhausted() || isLastUse || isBad)
		return false;
	return theUseLmt < 0 || theUseCnt < theUseLmt;
}

// prepare for next use if possible
// note: the notion of ``use'' is caller defined
void Connection::use() {
	Assert(reusable());
	// first user takes the hit of the opening delay
	theUseStart = theUseCnt++ ? (Time)TheClock : theOpenTime;
}
