
/*
 * $Id: Server.cc,v 1.44 1999/09/28 04:21:57 rousskov Exp $
 *
 *
 * Web Polygraph                        http://polygraph.ircache.net/
 * ------------------------------------------------------------------
 *
 *  Web Polygraph is copyrighted (C) 1999 by Alex Rousskov and the
 *  University of California, San Diego (UCSD); all rights reserved.
 *
 *  You may copy and distribute VERBATIM copies of this product. 
 *  You may modify your copy for your own use. You must NOT
 *  distribute modified copies. You must NOT publish benchmarking
 *  results based on modified versions.
 *
 *  If you publish any materials derived from or related to Web
 *  Polygraph, you must give a credit reference to the IRCache group.
 *
 *  Web Polygraph is distributed without any warranty, even implied.
 *
 *  Please see the COPYRIGHT file for full details.
 *
 */

#include "polygraph.h"

#include "Connection.h"
#include "RndPermut.h"
#include "ContentCfg.h"
#include "ContentMgr.h"
#include "Server.h"
#include "SrvConnMgr.h"
#include "ErrorMgr.h"
#include "StatPhaseMgr.h"
#include "LogComment.h"
#include "Rnd.h"

#include "ServerSym.h"
#include "ContentSym.h"
#include "SocketSym.h"

#include "polyErrors.h"
#include "gadgets.h"
#include "globals.h"

ObjFarm<SrvXact> Server::TheXacts;

static int TheSrvHistoryWin = 1000; // hard coded; how can we estimate this?


Server::Server(): theDirAccSel(0) {
	theConnMgr = new SrvConnMgr(this); // for now
}

Server::~Server() {
	delete theConnMgr;
	deaf();
}

void Server::configure(const ServerSym *cfg, bool bind2iface) {
	Assert(theConnMgr);

	SockOpt opt;
	Agent::configure(cfg, opt);

	configureContents(cfg);

	theHistory.window(TheSrvHistoryWin);

	// setup listen socket

	Must(theSock.create());
	Must(theSock.blocking(false));
	Should(theSock.reuseAddr(true));
	Must(theSock.configure(opt));

	if (bind2iface)
		Must(theSock.bind(theHost));
	else
		Must(theSock.bind(NetAddr(0, theHost.port())));

	Must(theSock.listen());

	int acceptLmt = -1;
	cfg->acceptLmt(acceptLmt);
	theConnMgr->configure(theHost, opt, cfg->pconnUseLmt());
	theConnMgr->acceptLmt(acceptLmt);

	TheXacts.limit(1024); // magic, no good way to estimate
}

void Server::start() {
	Assert(theSock);
	Assert(theConnMgr);
	Agent::start();
	theReserv = TheFileScanner->setFD(theSock.fd(), dirRead, this);
}

void Server::noteXactDone(SrvXact *x) {
	Assert(x);
	theConnMgr->put(x->conn());
	TheXacts.put(x);
}

void Server::noteReadReady(int) {
	if (!theConnMgr->accept(theSock)) {	// fatal error
		deaf();
		FatalError(errServerGone);
	}
}

void Server::noteConnReady(Connection *conn) {
	startXact(conn); // that's it?!
}

void Server::deaf() {
	if (theSock) {
		Comment(2) << "server " << theHost << " is closing listen socket " << theSock.fd() << endc;
		theSock.close();
	}
	if (theReserv)
		TheFileScanner->clearRes(theReserv);
}

void Server::startXact(Connection *conn) {
	SrvXact *x = TheXacts.get();
	x->exec(this, conn,
		theThinkDistr ? Time::Secd(theThinkDistr->trial()) : Time());
}

ContentCfg *Server::selectContent(const OidRec &or) {
	if (or.tid > 0)
		return TheContentMgr.get(or.tid);
	Assert(or.tid < 0);

	theDirAccSel->rndGen()->seed(RndPermut(or.oid, seqId()));
	const int tid = (int)theDirAccSel->trial();
	Assert(tid >= 0 && tid < theDirectAccess.count());
	return theDirectAccess[tid];
}

static
void mySyms2cfgs(Array<ContentSym*> &syms, Array<ContentCfg*> &cfgs) {
	for (int i = 0; i < syms.count(); ++i) {
		ContentCfg *cc = TheContentMgr.get(syms[i]);
		Check(cc->theSize);
		cfgs.append(cc);
	}
}

void Server::configureContents(const ServerSym *cfg) {
	Array<ContentCfg*> ccfgs;

	// get "global" content information
	Array<ContentSym*> csyms;
	Array<double> cprobs;
	if (!cfg->contents(csyms, cprobs) || !csyms.count()) {
		cerr << cfg->loc() << "no content specs for server " << cfg->kind() << endl;
		exit(-2);
	}
	Assert(csyms.count() == cprobs.count());
	mySyms2cfgs(csyms, ccfgs);

	// get direct access info
	Array<ContentSym*> dasyms;
	Array<double> daprobs;
	if (!cfg->directAccess(dasyms, theDirAccSel, daprobs) || !dasyms.count()) {
		cerr << cfg->loc() << "no direct_access specs for server " << cfg->kind() << endl;
		exit(-2);
	}
	Assert(theDirAccSel);
	mySyms2cfgs(dasyms, theDirectAccess);


	// check that all direct access cfgs are "global"
	for (int d = 0; d < theDirectAccess.count(); ++d) {
		ContentCfg *da = theDirectAccess[d];
		int count = 0;
		for (int c = 0; c < ccfgs.count(); ++c) {
			if (ccfgs[c]->id() == da->id())
				count++;
		}
		if (count != 1) {
			cerr << dasyms[d]->loc() 
				<< "direct_access spec below counted " << count
				<< " times in the global contents specs (must be 1):"
				<< endl << *dasyms[d] << endl;
			exit(-2);
		}
	}

	// give a chance to containers to compute setting that depend on
	// "global" content distribution
	for (int c = 0; c < ccfgs.count(); ++c) {
		ContentCfg *cc = ccfgs[c];

		// compute probability of producing an object with 'cc'
		double prod_prob = 0;
		for (int d = 0; d < theDirectAccess.count(); ++d) {
			ContentCfg *da = theDirectAccess[d];
			prod_prob += da->compLocalContrib(cc) * daprobs[d];
		}

		// check that all global content types are accessible
		if (prod_prob <= 0) {
			cerr << csyms[c]->loc()
				<< ": the content description below is not accessible "
				<< "(is not in direct_access closure):"
				<< endl << *csyms[c] << endl;
			exit(-2);
		}

		// notify contaners (and others) about prod_prob
		for (int d = 0; d < theDirectAccess.count(); ++d) {
			ContentCfg *da = theDirectAccess[d];
			da->noteGlobalContrib(cc, cprobs[c], prod_prob);
		}
	}	
}

